///////////////////////////////////////////////////////////////////////////////
// File    : interaction.cpp
// Content : interaction technique encapsulation
// History : 06-Mar-2006 - [rb] creation
//           19-Oct-2007 - [rb] added CeCIIL licence, prepared distro
///////////////////////////////////////////////////////////////////////////////

/* licence ********************************************************************
Copyright Renaud Blanch [http://iihm.imag.fr/blanch/]

This software is a computer program whose purpose is to explore new interaction
techniques for zoomable treemaps.

This software is governed by the CeCILL license under French law and abiding by 
the rules of distribution of free software. You can  use, modify and/ or 
redistribute the software under the terms of the CeCILL license as circulated  
at the following URL [http://www.cecill.info]. 

As a counterpart to the access to the source code and  rights to copy, modify 
and redistribute granted by the license, users are provided only with a limited 
warranty  and the software's author,  the holder of the economic rights,  and 
the successive licensors  have only limited liability. 

The fact that you are presently reading this means that you have had knowledge 
of the CeCILL license and that you accept its terms.
******************************************************************************/

/* includes ******************************************************************/

#include "interaction.h"

#include <iostream>
#include <cassert>
#include <cmath>
using namespace std;

#if defined(__APPLE__)
#	include <GLUT/glut.h>
#else
#	include <GL/glut.h>
#endif

#if !defined(GLUT_WHEEL_UP)
#	define GLUT_WHEEL_UP   3
#	define GLUT_WHEEL_DOWN 4
#endif

#include "colors.h"
#include "treemap.h"
#include "view.h"

#include "utils.h"


/* Interaction static data ***************************************************/

Interaction *Interaction::interaction = 0;
Interaction::Gesture Interaction::gesture;


/* Interaction implementation ************************************************/

Interaction::Interaction(const Treemap &t, View &v, const string &c) : 
	treemap(t), view(v), command(c),
	left(UP), middle(UP), right(UP), wheel(0),
	current(&treemap) {
	interaction = this;
	glPixelZoom(1., -1.);	
}

Interaction::~Interaction() {
	glutKeyboardFunc(0);
	glutKeyboardUpFunc(0);
	glutSpecialFunc(0);
	
	glutMouseFunc(0);
	glutMotionFunc(0);
	glutPassiveMotionFunc(0);
}

void Interaction::handleButton(int button, int state, int X, int Y) {
	bring_front();
	
	Xm = X;
	Ym = Y;
	
	switch(button) {
	case GLUT_LEFT_BUTTON:
		switch(state) {
				
		case GLUT_DOWN:
			Xl = X;
			Yl = Y;
			left = DOWN;
			glutPostRedisplay();
			break;
			
		case GLUT_UP:
			if(left == DOWN) {
				select_in(X, Y);
			} else {
				glutPostRedisplay();
			}
			left = UP;
			break;

		default: 
			assert(false); 
			break;
		}
			
		break;
		
	case GLUT_MIDDLE_BUTTON:
		switch(state) {
		case GLUT_DOWN:
			stroke_begin(X, Y);
			middle = DOWN;
			break;
			
		case GLUT_UP:
			middle = UP;
			stroke_end(X, Y);
			break;

		default:
			assert(false);
			break;
		}
			
		break;

	case GLUT_RIGHT_BUTTON:
		switch(state) {
		case GLUT_DOWN:
			Xr = X;
			Yr = Y;
			right = DOWN;
			glutPostRedisplay();
			break;
			
		case GLUT_UP:
			if(right == DOWN) {
				select_out(X, Y);
			} else {
				glutPostRedisplay();
			}
			right = UP;
			break;
			
		default: 
			assert(false); 
			break;
		}
		break;
	
	case GLUT_WHEEL_UP:
		wheel = 7;
		up(X, Y);
		break;
		
	case GLUT_WHEEL_DOWN:
		wheel = -7;
		down(X, Y);
		break;
		
	default:
		break;
	}
}


void Interaction::handleMotion(int X, int Y) {
	static const double DRAG_D = 5.;

	// pan
	switch(middle) {
	case UP: break;
	
	default:
		stroke(X, Y);
		break;
	}
	
	// stroke, select ...
	switch(left) {
	case UP: break;
			
	case DOWN:
		if(hypot(X - Xl, Y - Yl) > DRAG_D) {
			left = DRAG;
		}
		break;
		
	case DRAG:
		// dragging
		pan(X - Xm, Y - Ym);
		break;
	}
	
	// zoom
	switch(right) {
	case UP: break;
			
	case DOWN:
		if(hypot(X - Xr, Y - Yr) > DRAG_D) {
			right = DRAG;
		}
		break;
		
	case DRAG:
		zoom(Xr, Yr, (X - Xm) - (Y - Ym));
		break;
	}
	
	Xm = X;
	Ym = Y;
}


/* feedback ******************************************************************/

#define rect(r, GL_TYPE) {  \
	glBegin(GL_TYPE);       \
	glVertex2f(r.x0, r.y0); \
	glVertex2f(r.x0, r.y1); \
	glVertex2f(r.x1, r.y1); \
	glVertex2f(r.x1, r.y0); \
	glEnd();                \
}

#define txtl(s, X, Y) { \
	glRasterPos2f(View::x(X), View::y(Y)); \
	for(const char *c = s.c_str(); *c != 0; ++c) { \
		glutBitmapCharacter(GLUT_BITMAP_9_BY_15, int(*c)); \
	} \
}

#define txtr(s, X, Y) { \
	int Xs = X - 9 * s.size(); \
	double y = view.y(Y); \
	GLboolean valid; \
	glRasterPos2f(view.x(Xs), y); \
	glGetBooleanv(GL_CURRENT_RASTER_POSITION_VALID, &valid); \
	const char *c = s.c_str(); \
	while(not valid and *c != 0) { \
		Xs += 9; \
		glRasterPos2f(view.x(Xs), y); \
		glGetBooleanv(GL_CURRENT_RASTER_POSITION_VALID, &valid); \
		++c; \
	} \
	for(; *c != 0; ++c) { \
		glutBitmapCharacter(GLUT_BITMAP_9_BY_15, int(*c)); \
	} \
}

#define pict(name, X, Y) { \
	glRasterPos2f(view.x(X - name##_x_hot), \
	              view.y(Y - name##_y_hot)); \
	glDrawPixels(name##_width, name##_height, \
	             GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, \
	             name##_luminance_alpha); \
}


void Interaction::renderFeedback() const {
	glPushMatrix();
	glTranslatef(0, 0, view.box.depth() - 1);	

//	// margin
//	Rectangle target = view.target();
//	Rectangle mask(-1, -1, 2, 2);
//	glColor4f(.5, .5, .5, .7);
//	mask.x1 = target.x0;
//	rect(mask, GL_QUADS);
//	mask.x0 = target.x1; mask.x1 = 2.;
//	rect(mask, GL_QUADS);
//	mask.x0 = target.x0; mask.x1 = target.x1;
//	mask.y1 = target.y0;
//	rect(mask, GL_QUADS);
//	mask.y0 = target.y1; mask.y1 = 2.;
//	rect(mask, GL_QUADS);
	
	// current feedback
	glColor3fv(colors::black);
	txtr(current->path(), view.W() - 5, 15);
	rect(current->box, GL_LINE_LOOP);

	glColor4f(0.5, 0.5, 0.5, .1);
	Rectangle mask;
	mask.x1 = current->box.x0;
	rect(mask, GL_QUADS);
	mask.x0 = current->box.x1; mask.x1 = 1.;
	rect(mask, GL_QUADS);
	mask.x0 = current->box.x0; mask.x1 = current->box.x1;
	mask.y1 = current->box.y0;
	rect(mask, GL_QUADS);
	mask.y0 = current->box.y1; mask.y1 = 1.;
	rect(mask, GL_QUADS);

	if(!gesture.empty()) {
		// stroke selection
		double animation = View::getAnimationState();
		if(animation == 0. && middle == UP) {
			gesture.clear();
					
		} else {
			// stroke
			glLineWidth(3);
			glColor4f(.5, .5, 1, (animation == 0.? 1. : animation));
			glBegin(GL_LINE_STRIP);
			double x, y;
			for(Gesture::const_iterator i = gesture.begin(); i != gesture.end();) {
				x = *i++; y = *i++;
				glVertex2d(x, y);
			}
			glEnd();

			// selection
			rect(selection->box, GL_LINE_LOOP);
			glLineWidth(1);
			glColor4f(0.5, 0.5, 1, .1);
			rect(selection->box, GL_QUADS);
			glColor3f(0, 0, 0);
			txtr(selection->path(), Xm - 8, Ym + 22);
		}
	}

	// mouse state feedback
#define MOUSE_FEEDBACK
#if defined(MOUSE_FEEDBACK)
	{

#		include "../img/zoom.h"
#		include "../img/mouse.h"
#		include "../img/left.h"
#		include "../img/middle.h"
#		include "../img/right.h"
#		include "../img/up.h"
#		include "../img/down.h"
	
		if(right == DRAG)  pict(zoom, Xr, Yr);
	
		// mouse proxy
		static const int Xc = 12;
		static const int Yc = -1; //14;
		if(left != UP or middle != UP or right != UP or wheel != 0) {
			pict(mouse, Xm + Xc, Ym + Yc);
			if(left   != UP) pict(left,   Xm + Xc, Ym + Yc);
			if(middle != UP) pict(middle, Xm + Xc, Ym + Yc);
			if(right  != UP) pict(right,  Xm + Xc, Ym + Yc);
			if(wheel != 0) {
				if(wheel > 0) {
					pict(up, Xm + Xc, Ym + Yc);
					wheel -= 1;
				} else {
					pict(down, Xm + Xc, Ym + Yc);
					wheel += 1;
				}
				glutPostRedisplay();
			}
		}
	}
#endif // defined(MOUSE_FEEDBACK)

	glPopMatrix();
}

#undef rect
#undef text
#undef pict


/* selection *****************************************************************/

void Interaction::select_in(int X, int Y) {
	double x = view.x(X);
	double y = view.y(Y);
	
	while(not current->box.contains(x, y) && current != &treemap) {
		current = current->up();
	}

	if(not glutGetModifiers() & GLUT_ACTIVE_SHIFT) {
		current = current->down(x, y);
		view.zoom(current->box);
	} else {
		open(command.c_str(), current->path().c_str());
		glutPostRedisplay();
	}

}

void Interaction::select_out(int, int) {
	current = current->up();
	view.zoom(current->box);
}


/* pan & zoom ****************************************************************/

void Interaction::pan(int dX, int dY) {
	view.pan(-dX * view.sx(), dY * view.sy());
	glutPostRedisplay();
}

void Interaction::zoom(int X, int Y, double dz /* = 0 */) {
	const double ds2 = exp(dz * 0.02);
	
	const double x = view.x(X);
	const double y = view.y(Y);
	
	current = &treemap;
	selection = current->down(x, y);
	Rectangle target = view.target();
	const double size_limit = target.size();
	while(selection->box.size() * ds2 >= size_limit and selection != current) {
		current = selection;
		selection = current->down(x, y);
	}
	
	if(ds2 < 1) {
		selection = current;
	}

	const double Dsx = target.w() / selection->box.w();
	const double Dsy = target.h() / selection->box.h();
	const double t = log(ds2) / log(Dsx * Dsy);
	view.zoom(pow(Dsx, t), pow(Dsy, t), x, y);
	
	glutPostRedisplay();
}


/* cross selection ***********************************************************/

void Interaction::stroke_begin(int X, int Y) {
	double x = view.x(X);
	double y = view.y(Y);
	gesture.clear();
	gesture.push_back(x);
	gesture.push_back(y);
	selection = treemap.pick(x, y);
	
	glutPostRedisplay();
}

void Interaction::stroke(int X, int Y) {
	double x = view.x(X);
	double y = view.y(Y);
	gesture.push_back(x);
	gesture.push_back(y);
	while(not selection->box.contains(x, y) && selection != &treemap) {
		selection = selection->up();
	}
	
	glutPostRedisplay();
}

void Interaction::stroke_end(int X, int Y) {
	if(not glutGetModifiers() & GLUT_ACTIVE_SHIFT) {
		// normal interaction: zoom to selected node
		current = selection;
		view.zoom(current->box);
	} else {
		open(command.c_str(), selection->path().c_str());
		glutPostRedisplay();
	}
}
