///////////////////////////////////////////////////////////////////////////////
// File    : placer.cpp
// Content : treemap placer
// History : 10-Apr-2006 - [rb] creation
//           19-Oct-2007 - [rb] added CeCIIL licence, prepared distro
///////////////////////////////////////////////////////////////////////////////

/* licence ********************************************************************
Copyright Renaud Blanch [http://iihm.imag.fr/blanch/]

This software is a computer program whose purpose is to explore new interaction
techniques for zoomable treemaps.

This software is governed by the CeCILL license under French law and abiding by 
the rules of distribution of free software. You can  use, modify and/ or 
redistribute the software under the terms of the CeCILL license as circulated  
at the following URL [http://www.cecill.info]. 

As a counterpart to the access to the source code and  rights to copy, modify 
and redistribute granted by the license, users are provided only with a limited 
warranty  and the software's author,  the holder of the economic rights,  and 
the successive licensors  have only limited liability. 

The fact that you are presently reading this means that you have had knowledge 
of the CeCILL license and that you accept its terms.
******************************************************************************/

/* includes ******************************************************************/

#include "placer.h"

#include <algorithm>
#include <iostream>
#include <cassert>
using namespace std;


/* Placer factory ************************************************************/

Placer *Placer::create(char type) {
	switch(type) {
	case SQUARIFIED:
		return new Squarified();
		break;
		
	case SLICE_AND_DICE:
		return new SliceAndDice();
		break;
		
	default:
		assert(false);
		break;
	}
}


/* SliceAndDice implementation ***********************************************/

/*
	Original slice-and-dice strategy.
	The global rectangle is sliced alternativelly horizontally and vertically.
	
	Shneiderman, B. (1992).
	Tree visualization with tree-maps: 2-d space-filling approach.
	ACM Trans. Graph., 11(1):92--99.
*/

RectangleList SliceAndDice::split(const Rectangle &box, 
                                  int level, double weight,
                                  const WeightList &weights) const {
	if(count++ % 1000 == 0) cout << "." << flush;

	RectangleList rectangles;
	Rectangle subbox = box;
	WeightList::const_iterator e = weights.end();
	if(level % 2 != 0) {
		Rectangle::value_type wpw = box.w() / weight;
		for(WeightList::const_iterator w = weights.begin(); w != e; ++w) {
			subbox.x1 = subbox.x0 + (*w) * wpw;
			rectangles.push_back(subbox);
			subbox.x0 = subbox.x1;
		}
	} else {
		Rectangle::value_type hpw = box.h() / weight;
		for(WeightList::const_iterator w = weights.begin(); w != e; ++w) {
			subbox.y0 = subbox.y1 - (*w) * hpw;
			rectangles.push_back(subbox);
			subbox.y1 = subbox.y0;
		}	
	}
	return rectangles;
}


/* Squarified implementation *************************************************/

/*
	Squarified strategy, the order is not preserved but the subrectangles
	aspect ratio are closer to 1.
	
	Bruls, M., Huizing, K., and van Wijk, J.J. (2000). 
	Squarified treemaps. 
	In Proc. Eurographics & IEEE TCVG Symp. on Visualization, pages 33--42
*/


// predicate used to sort an array according to another container order
template< class C >
class IndexedGreaterThan {
	typedef typename C::const_iterator RandomAccessIterator;
	typedef typename C::size_type      size_type;
	
	RandomAccessIterator first;
public:
	IndexedGreaterThan(const RandomAccessIterator &i) : first(i) {}
	bool operator()(size_type a, size_type b) const {
		return *(first + a) > *(first + b);
	}
};


RectangleList Squarified::split(const Rectangle &box, 
                                int level, double weight,
                                const WeightList &weights) const {
	if(count++ % 1000 == 0) cout << "." << flush;

	// building order with increasing weights
	size_t size = weights.size();
	size_t *order = new size_t[size];
	for(size_t i = 0; i < size; ++i) { order[i] = i; }
	sort(order, order + size, 
	     IndexedGreaterThan< WeightList >(weights.begin()));

	// building rectangle list
	RectangleList rectangles(size);

	Rectangle empty = box;
	double weight_remaining = weight;
	for(size_t stop, start = 0; start < size;) {
		double r = empty.w() / empty.h();
		const double er = max(r, 1/r);
		
		double weight_placed = weights[order[start]];
		double p = weight_placed / weight_remaining;
		r = er*p;
		double max_r = max(r, 1/r);
		
		// placing the biggest children till their "squarinness" degrades
		for(stop = start + 1; stop < size; ++stop) {
			weight_placed += weights[order[stop]];
			p = (weight_placed * weight_placed * er) / weight_remaining;
			r = max(weights[order[start]]/p, p/weights[order[stop]]);
			if(r > max_r) {
				weight_placed -= weights[order[stop]];
				break;
			}
			max_r = r;
		}
		
		if(empty.w() >= empty.h()) {
			// vertical slice
			Rectangle subbox = empty;
			subbox.x1 = subbox.x0 + subbox.w() * weight_placed / weight_remaining;
			double hpw = subbox.h() / weight_placed;
			for(; start < stop; ++start) {
				subbox.y0 = subbox.y1 - weights[order[start]] * hpw;
				rectangles[order[start]] = subbox;
				subbox.y1 = subbox.y0;
			}
			empty.x0 += subbox.w();
			
		} else {
			// horizontal slice
			Rectangle subbox = empty;
			subbox.y0 = subbox.y1 - subbox.h() * weight_placed / weight_remaining;
			double wpw = subbox.w() / weight_placed;
			for(; start < stop; ++start) {
				subbox.x1 = subbox.x0 + weights[order[start]] * wpw;
				rectangles[order[start]] = subbox;
				subbox.x0 = subbox.x1;
			}
			empty.y1 -= subbox.h();
		}
		
		weight_remaining -= weight_placed;
	}

	delete[] order;
	return rectangles;
}
