///////////////////////////////////////////////////////////////////////////////
// File    : renderer.h
// Content : treemap renderer
// History : 10-Apr-2006 - [rb] creation
//           19-Oct-2007 - [rb] added CeCIIL licence, prepared distro
///////////////////////////////////////////////////////////////////////////////

/* licence ********************************************************************
Copyright Renaud Blanch [http://iihm.imag.fr/blanch/]

This software is a computer program whose purpose is to explore new interaction
techniques for zoomable treemaps.

This software is governed by the CeCILL license under French law and abiding by 
the rules of distribution of free software. You can  use, modify and/ or 
redistribute the software under the terms of the CeCILL license as circulated  
at the following URL [http://www.cecill.info]. 

As a counterpart to the access to the source code and  rights to copy, modify 
and redistribute granted by the license, users are provided only with a limited 
warranty  and the software's author,  the holder of the economic rights,  and 
the successive licensors  have only limited liability. 

The fact that you are presently reading this means that you have had knowledge 
of the CeCILL license and that you accept its terms.
******************************************************************************/

#if !defined(RENDERER_H)
#define RENDERER_H


/* includes ******************************************************************/

#include <string>
#include <ostream>


/* forward declaration *******************************************************/

class Rectangle;
class View;


/* Abstract Renderer *********************************************************/

class Renderer {
protected:
	static double depth_max;
	static int start_time;	
	static double ptx, pty;

protected:
	const View *view;
	const int text_size_min;
	const int text_size_max;
	const float alpha;
	const int palette;

public:
	Renderer(const View *view = 0, 
	         int text_size_min = 10, int text_size_max = 100,
	         float alpha = .3, int palette = 0);
	virtual ~Renderer() {}
	
	static Renderer *create(char q, char t, 
	                        const View *view, 
	                        int text_size_min, int text_size_max, 
	                        float alpha, int palette);

	// tree traversal
	void preamble() const;
	void postamble() const;

	bool top_down (const Rectangle &box) const;
	void bottom_up(const Rectangle &box, int level, 
	               const std::string &label) const;

protected:	
	// abstract rendering
	virtual void quad(const Rectangle &box, int level) const = 0;
	virtual void text(const Rectangle &box, int level, 
	                  const std::string &label) const = 0;
};


/* Concrete Graphic Renderer Mixins ******************************************/

struct OutlineQuad : virtual Renderer {
	virtual void quad(const Rectangle &box, int level) const;
};

struct RampQuad : virtual Renderer {
	RampQuad();
	virtual void quad(const Rectangle &box, int level) const;
};


struct HorizontalText : virtual Renderer {
	virtual void text(const Rectangle &box, int level, 
	                  const std::string &label) const;
};

struct CenteredText : virtual Renderer {
	virtual void text(const Rectangle &box, int level, 
	                  const std::string &label) const;
};

struct OrthogonalText : virtual Renderer {
	virtual void text(const Rectangle &box, int level, 
	                  const std::string &label) const;
};


/* Composite Concrete Renderers **********************************************/

template< class Q, class T >
class CRenderer : public virtual Renderer, protected Q, protected T {
public:
	CRenderer(const View *v, int t_min, int t_max, float a, int p) :
		Renderer(v, t_min, t_max, a, p) {}
protected:
	virtual void quad(const Rectangle &box, int level) const {
		Q::quad(box, level); }
	virtual void text(const Rectangle &box, int level, 
	                  const std::string &label) const {
		T::text(box, level, label); }
};


#endif // !defined(RENDERER_H)

