///////////////////////////////////////////////////////////////////////////////
// File    : treemap.cpp
// Content : basic tree structure
// History : 10-Apr-2006 - [rb] creation
//           19-Oct-2007 - [rb] added CeCIIL licence, prepared distro
///////////////////////////////////////////////////////////////////////////////

/* licence ********************************************************************
Copyright Renaud Blanch [http://iihm.imag.fr/blanch/]

This software is a computer program whose purpose is to explore new interaction
techniques for zoomable treemaps.

This software is governed by the CeCILL license under French law and abiding by 
the rules of distribution of free software. You can  use, modify and/ or 
redistribute the software under the terms of the CeCILL license as circulated  
at the following URL [http://www.cecill.info]. 

As a counterpart to the access to the source code and  rights to copy, modify 
and redistribute granted by the license, users are provided only with a limited 
warranty  and the software's author,  the holder of the economic rights,  and 
the successive licensors  have only limited liability. 

The fact that you are presently reading this means that you have had knowledge 
of the CeCILL license and that you accept its terms.
******************************************************************************/

/* includes ******************************************************************/

#include "treemap.h"

#include <iostream>
#include <cassert>
#include <algorithm>
#include <limits>
using namespace std;

#include "placer.h"
#include "renderer.h"


/* destruction & construction ************************************************/

Treemap::~Treemap() {
	ChildMap::iterator e = children.end();
	for(ChildMap::iterator child = children.begin(); child != e; ++child) {
		delete child->second;
	}
	children.clear();
}

Treemap::Treemap(istream &input) : 
	Rectangle(), parent(this), weight(0.), level(0), label(""), box(*this) {
	// loading
	cout << "reading input ('.' for 1k)" << endl;
	int count = 0;
	
	string node_path;
	double node_weight;
	while(!input.eof()) {
		input >> node_path >> node_weight;
		
		// handling malformed line (by ignoring it)
		if(!input) {
			input.clear();
			input.ignore(numeric_limits< int >::max(), '\n');
			continue;
		}
		
		// skipping empty nodes
		if(node_weight <= 0.) continue;	
		
		count += 1;
		if(count % 1000 == 0) cout << '.' << flush;
		
		replace(node_path.begin(), node_path.end(), '_', ' ');
		append(node_path, node_weight);
	}

	cout << " done" << endl;
	cout << count << " nodes loaded (total weight: " << weight << ")" << endl;
}

void Treemap::append(std::string node_path, double node_weight) {
	assert(node_weight > 0.);
	
	weight += node_weight;
	
	// terminal condition
	if(node_path == "") return;

	// splitting path
	string node_label;
	string::size_type sep = node_path.find('/');
	if(sep != string::npos) {
		node_label = string(node_path.begin(), node_path.begin() + sep);
		node_path  = string(node_path.begin() + sep + 1, node_path.end());
	} else {
		node_label = node_path;
		node_path  = "";
	}
		
	// finding appropriate subnode
	ChildMap::iterator child = children.find(node_label);
	Treemap *node = 0;
	if(child == children.end()) {
		node = children[node_label] = new Treemap(this, node_label);
	} else {
		node = child->second;
	}
	node->append(node_path, node_weight);	
}


/* navigation ****************************************************************/

const Treemap *Treemap::down(double x, double y) const {
	const ChildMap::const_iterator e = children.end();
	for(ChildMap::const_iterator c = children.begin(); c != e; ++c) {
		const Treemap *child = c->second;
		if(child->contains(x, y)) {
			return child;
		}
	}
	
	return this;
}

const Treemap *Treemap::pick(double x, double y) const {
	const Treemap *result = this;
	const Treemap *d = 0;
	while((d = result->down(x, y)) != result) {
		result = d;
	}
	return result;
}

std::string Treemap::path() const {
	return (parent == this ? "" : parent->path() + "/") + label;
}


/* layout & display **********************************************************/

void Treemap::place(const Placer *placer) {
	cout << "computing layout" << endl;
	do_place(placer); 
	cout << " done" << endl;
	cout << placer->count << " nodes placed" << endl;
}

void Treemap::render(const Renderer *renderer) const {
	renderer->preamble();
	do_render(renderer);
	renderer->postamble();
}


void Treemap::do_place(const Placer *placer) {
	WeightList weights;
	const ChildMap::const_iterator e = children.end();
	for(ChildMap::const_iterator c = children.begin(); c != e; ++c) {
		weights.push_back(c->second->weight);
	}
	
	RectangleList rectangles = placer->split(box, level, weight, weights);
	
	RectangleList::const_iterator r = rectangles.begin();
	for(ChildMap::const_iterator c = children.begin(); c != e; ++c) {
		Treemap *child = c->second;
		Rectangle *rectangle = child;
		*rectangle = *r++;
		child->do_place(placer);
	}
}

void Treemap::do_render(const Renderer *renderer) const {
	if(renderer->top_down(box)) {
		const ChildMap::const_iterator e = children.end();
		for(ChildMap::const_iterator c = children.begin(); c != e; ++c) {
			c->second->do_render(renderer);
		}
		renderer->bottom_up(box, level, label);
	}
}


/* output ********************************************************************/

void Treemap::print(ostream &os) const {
	os << "<?xml version=\"1.0\" standalone=\"no\"?>" << endl
	   << "<!DOCTYPE tree SYSTEM \"http://www.cs.umd.edu/hcil/iv03contest/datasets/treeml.dtd\">" << endl
	   << "<tree>" << endl
	   << "	<declarations>" << endl
	   << "		<attributeDecl name=\"name\" type=\"String\"/>" << endl
	   << "	</declarations>" << endl;
	
//	os << "<?xml version=\"1.0\" standalone=\"no\"?>" << endl
//	   << "<tree>" << endl;
	
	do_print(os);
	
	os << "</tree>" << endl;
}

void Treemap::do_print(ostream &os) const {
	string indent(level + 1, '\t');
	
	if(children.empty()) {
		// leaf
		os << indent << "<leaf>" << endl 
		   << indent << "\t<attribute name=\"name\" value=\"" << label << "\"/>" << endl 
		   << indent << "</leaf>" << endl;

//		os << indent << "<leaf name=\"" << label << "\"/>" << endl;
		   
	} else {
		// branch
		os << indent << "<branch>" << endl 
		   << indent << "\t<attribute name=\"name\" value=\"" << label << "\"/>" << endl;

//		os << indent << "<branch name=\"" << label << "\">" << endl;
		
		const ChildMap::const_iterator e = children.end();
		for(ChildMap::const_iterator c = children.begin(); c != e; ++c) {
			const Treemap *child = c->second;
			child->do_print(os);
		}
		
		os << indent << "</branch>" << endl;
		
	}
}

