///////////////////////////////////////////////////////////////////////////////
// File    : ztm.cpp
// Content : my second treemap implementation
// History : 10-Apr-2006 - [rb] recreation
//           19-Oct-2007 - [rb] added CeCIIL licence, prepared distro
///////////////////////////////////////////////////////////////////////////////

/* licence ********************************************************************
Copyright Renaud Blanch [http://iihm.imag.fr/blanch/]

This software is a computer program whose purpose is to explore new interaction
techniques for zoomable treemaps.

This software is governed by the CeCILL license under French law and abiding by 
the rules of distribution of free software. You can  use, modify and/ or 
redistribute the software under the terms of the CeCILL license as circulated  
at the following URL [http://www.cecill.info]. 

As a counterpart to the access to the source code and  rights to copy, modify 
and redistribute granted by the license, users are provided only with a limited 
warranty  and the software's author,  the holder of the economic rights,  and 
the successive licensors  have only limited liability. 

The fact that you are presently reading this means that you have had knowledge 
of the CeCILL license and that you accept its terms.
******************************************************************************/

/* includes ******************************************************************/

#include <cstdlib>
#include <iostream>
using namespace std;

#if defined(__APPLE__)
#	include <GLUT/glut.h>
#else
#	include <GL/glut.h>
#endif

#include "placer.h"
#include "renderer.h"
#include "treemap.h"
#include "view.h"
#include "interaction.h"


/* declarations **************************************************************/

void reshape(int W, int H);
void display();
void keyboard(unsigned char key, int X, int Y);
void keyup(unsigned char key, int X, int Y);
void special(int key, int X, int Y);

void print_help();


/* state *********************************************************************/

struct State {
	Placer *placer;
	Renderer *renderer;
	
	Treemap treemap;
	View view;
	Interaction interaction;
	
	State(int W, int H, int MARGIN, 
	      int step_min, int step_max, string command) :
		placer(0), renderer(0),
		treemap(cin),
		view(W, H, MARGIN, step_min, step_max), 
		interaction(treemap, view, command) {
	}
	~State() {
		delete placer;
		delete renderer;
	}
};

State *state = 0;


/* usage *********************************************************************/

void clean() {
	if(state != 0) delete state;
	cout << "bye." << endl;
}

void exit_usage(const char *name) {
	cerr << "usage: " << name << " [-cawhxymtTsSpflrv]" << endl;
	cerr << "-c <command> script to be executed." << endl;
	cerr << "-a <alpha>   attenuation between levels." << endl;
	cerr << "-w <width>   window width." << endl;
	cerr << "-h <height>  window height." << endl;
	cerr << "-x <x>       window x." << endl;
	cerr << "-y <y>       window y." << endl;
	cerr << "-m <size>    margins size." << endl;
	cerr << "-e <size>    minimal text size." << endl;
	cerr << "-E <size>    minimal text size." << endl;
	cerr << "-s <n>       minimal animation steps." << endl;
	cerr << "-S <n>       maximal animation steps." << endl;
	cerr << "-p <n>       palette." << endl;
	cerr << "-f           full screen." << endl;
	cerr << "-l [d|q]     layout algorithm." << endl;
	cerr << "             slice&dice/squarified" << endl;
	cerr << "-q [o|r]     quad renderer." << endl;
	cerr << "             outline/ramp fill" << endl;
	cerr << "-t [o|h|c]   text renderer." << endl;
	cerr << "             orthogonal/horizontal/centered" << endl;
	cerr << "-v           verbose mode." << endl;
	exit(EXIT_FAILURE);
}


/* main **********************************************************************/

int main(int argc, char *argv[]) {
	// initialisation
	atexit(clean);
	glutInit(&argc, argv);
	
	// command line
	const char *name = argv[0];
	cout << name << " - [rb] (c) 2006" << endl;
	
	string command = "../scripts/echo.sh";
	
	float alpha = .3;
	int W = 640;
	int H = 480;
	int X = 100;
	int Y = 100;
	int margin = 15;
	int text_size_min = 20;
	int text_size_max = 100;
	int step_min = 15;
	int step_max = 15;
	int palette = 0;
	bool full_screen = false;

	char placer = 'd'; // sliceandDice
	char quad   = 'o'; // Outline
	char text   = 'o'; // Orthogonal
	
	bool verbose = false;
	
	for(int i = 1; i < argc;) {
		const char *arg = argv[i++];
		if(arg[0] != '-') exit_usage(name);
		
		switch(arg[1]) {

		#define setv(c, v, t) \
		case c: \
			if(i + 1 > argc) exit_usage(name); \
			v = ato##t(argv[i++]); \
			break; 

		#define setc(c, v) \
		case c: \
			if(i + 1 > argc) exit_usage(name); \
			v = argv[i++][0]; \
			break; 

		#define sets(c, v) \
		case c: \
			if(i + 1 > argc) exit_usage(name); \
			v = argv[i++]; \
			break;

		#define setb(c, v) \
		case c: \
			v = true; \
			break;

		sets('c', command);	
		setv('a', alpha, f);
		setv('w', W, i);
		setv('h', H, i);
		setv('x', X, i);
		setv('y', Y, i);
		setv('m', margin, i);
		setv('e', text_size_min, i);
		setv('E', text_size_max, i);
		setv('s', step_min, i);
		setv('S', step_max, i);
		setv('p', palette, i);
		setb('f', full_screen);
		setc('l', placer);
		setc('q', quad);
		setc('t', text);
		setb('v', verbose);
		
		#undef setv
		#undef setc
		#undef sets
		#undef setb

		
		default:
			exit_usage(name);
			break;
		}
	}
	
	if(verbose) {
		cout << "configuration:" << endl;

		#define print(v) cout << #v << " = " << v << endl
		print(command);
		print(alpha);
		print(W);
		print(H);
		print(X);
		print(Y);
		print(margin);
		print(text_size_min);
		print(text_size_max);
		print(step_min);
		print(step_max);
		print(palette);
		print(full_screen);
		print(placer);
		print(quad);
		print(text);
		print(verbose);
		#undef print
	}
	
	// window
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
	glutInitWindowSize(W, H);
	glutInitWindowPosition(X, Y);
	glutCreateWindow(name);
	if(full_screen) glutFullScreen();

	// treemap
	state = new State(W, H, margin, step_min, step_max, command);
	state->placer = Placer::create(placer);
	state->treemap.place(state->placer);
	state->renderer = Renderer::create(quad, text, &state->view, 
	                                   text_size_min, text_size_max,
	                                   alpha, palette);
	
	// opengl projection
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, 1, 0, 1, 0, 50);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// interaction
	glutReshapeFunc(reshape);
	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyup);
	glutSpecialFunc(special);
	
	glutMouseFunc(Interaction::mouse);
	glutMotionFunc(Interaction::motion);
	glutPassiveMotionFunc(Interaction::motion);
	
	// go !
	glutMainLoop();
	
	return EXIT_SUCCESS;
}


/* display *******************************************************************/

void reshape(int W, int H) {
	glViewport(0, 0, W, H);
	state->view.reshape(W, H);
}

void display() {
	glClear(GL_COLOR_BUFFER_BIT);

	glPushMatrix();
	state->view.setGLMatrix();
	state->treemap.render(state->renderer);	
	state->interaction.renderFeedback();
	glPopMatrix();
	
	glutSwapBuffers();
}


/* keyboard interaction ******************************************************/
	
void keyboard(unsigned char key, int X, int Y) {
	static const double SCALE_U = 1.1;
	static const double SCALE_D = 1./SCALE_U;

	static int W = glutGet(GLUT_WINDOW_WIDTH);
	static int H = glutGet(GLUT_WINDOW_HEIGHT);
	
	switch(key) {
	case 'p':
		cout << state->treemap << endl;
		break;
		
	// small (pda-like) screen-size
	case 's':
		glutReshapeWindow(W, H);
		break;
	
	case 'f':
		W = glutGet(GLUT_WINDOW_WIDTH);
		H = glutGet(GLUT_WINDOW_HEIGHT);
		glutFullScreen();
		break;
	

	// keyboard interaction
	case 13: // return
		state->interaction.select(X, Y);
		break;
		
	case 127: // backspace
		state->interaction.back(X, Y);
		break;

	// zoom	
	case 'y':
		// mouse centered 
		state->view.zoom(SCALE_U, SCALE_U, state->view.x(X), state->view.y(Y));
		break;

	case 'h':
		// view centered
		state->view.zoom(SCALE_D, SCALE_D);
		break;

		
	case 'u': state->view.zoom(SCALE_U, 1., state->view.x(X), 0.); break;	
	case 'j': state->view.zoom(SCALE_D, 1.);                       break;	

	case 'i': state->view.zoom(1., SCALE_U, 0., state->view.y(Y)); break;
	case 'k': state->view.zoom(1., SCALE_D);                       break;
		
	// space pseudo button (see keyup)
	case ' ':
		glutIgnoreKeyRepeat(true);
		Interaction::mouse(GLUT_LEFT_BUTTON, GLUT_DOWN, X, Y);
		break;

		
	// quit	
	case 'q':
	case  27:   // esc
		exit(EXIT_SUCCESS);
		break;

	// help !
	default:
		print_help();
		break;
	}
	glutPostRedisplay();
}

void keyup(unsigned char key, int X, int Y) {
	switch(key) {
		case ' ':
			glutIgnoreKeyRepeat(false);
			Interaction::mouse(GLUT_LEFT_BUTTON, GLUT_UP, X, Y);
			break;
			
		default:
			break;
	}
}

void special(int key, int X, int Y) {
	switch(key) {
	case GLUT_KEY_PAGE_UP:
		state->interaction.up(X, Y);
		break;
		
	case GLUT_KEY_PAGE_DOWN:
		state->interaction.down(X, Y);
		break;
		
	default:
		break;
	}
	glutPostRedisplay();
}


/* help **********************************************************************/

void print_help() {
	cout << "keyboard commands:" << endl;
	cout << "s/f: small/full screen" << endl;
	cout << "return:    enter level" << endl;
	cout << "backspace: leave level" << endl;
	cout << "y/h: x & y (de)zoom" << endl;
	cout << "u/j: x (de)zoom" << endl;
	cout << "i/k: y (de)zoom" << endl;
	cout << "q, esc: quit" << endl;
}
